#----------------------------------#    ---- NOTES ----
# File name: TachoMod_V4_0.py      #    V1 = first working version, without display (fastest processing!)
# Author: cosmowave                #    V2 = version with display and menu (in higher subversions) for changing settings 
# Date created: 6.9.2022           #    V3 = version which automatically stores settings in a file
# Date last modified: 13.10.2022   #    V4 = implemented pulse per revolution for correct rpm calculation
# Version 4.0                      #
#----------------------------------#

# ---- imports ----
from machine import Pin, SPI
import time
import framebuf
import _thread
import os

# ---- OLED DEFINITION----
DC = 8
RST = 12
MOSI = 11
SCK = 10
CS = 9

class OLED_1inch3(framebuf.FrameBuffer):
    def __init__(self):
        self.width = 128
        self.height = 64

        self.cs = Pin(CS,Pin.OUT)
        self.rst = Pin(RST,Pin.OUT)

        self.cs(1)
        self.spi = SPI(1)
        self.spi = SPI(1,2000_000)
        self.spi = SPI(1,20000_000,polarity=0, phase=0,sck=Pin(SCK),mosi=Pin(MOSI),miso=None)
        self.dc = Pin(DC,Pin.OUT)
        self.dc(1)
        self.buffer = bytearray(self.height * self.width // 8)
        super().__init__(self.buffer, self.width, self.height, framebuf.MONO_HMSB)
        self.init_display()
        
        self.white =   0xffff
        self.balck =   0x0000
        
    def write_cmd(self, cmd):
        self.cs(1)
        self.dc(0)
        self.cs(0)
        self.spi.write(bytearray([cmd]))
        self.cs(1)

    def write_data(self, buf):
        self.cs(1)
        self.dc(1)
        self.cs(0)
        self.spi.write(bytearray([buf]))
        self.cs(1)

    def init_display(self):
        """Initialize dispaly"""  
        self.rst(1)
        time.sleep(0.001)
        self.rst(0)
        time.sleep(0.01)
        self.rst(1)
        
        self.write_cmd(0xAE)    # turn off OLED display

        self.write_cmd(0x00)    # set lower column address
        self.write_cmd(0x10)    # set higher column address 

        self.write_cmd(0xB0)    # set page address 
      
        self.write_cmd(0xdc)    # et display start line 
        self.write_cmd(0x00) 
        self.write_cmd(0x81)    # contract control 
        self.write_cmd(0x6f)    # 128
        self.write_cmd(0x21)    # Set Memory addressing mode (0x20/0x21) #
    
        self.write_cmd(0xa0)    # set segment remap 
        self.write_cmd(0xc0)    # Com scan direction
        self.write_cmd(0xa4)    # Disable Entire Display On (0xA4/0xA5) 

        self.write_cmd(0xa6)    # normal / reverse
        self.write_cmd(0xa8)    # multiplex ratio 
        self.write_cmd(0x3f)    # duty = 1/64
  
        self.write_cmd(0xd3)    # set display offset 
        self.write_cmd(0x60)

        self.write_cmd(0xd5)    # set osc division 
        self.write_cmd(0x41)
    
        self.write_cmd(0xd9)    # set pre-charge period
        self.write_cmd(0x22)   

        self.write_cmd(0xdb)    # set vcomh 
        self.write_cmd(0x35)  
    
        self.write_cmd(0xad)    # set charge pump enable 
        self.write_cmd(0x8a)    # Set DC-DC enable (a=0:disable; a=1:enable)
        self.write_cmd(0XAF)
    def show(self):
        self.write_cmd(0xb0)
        for page in range(0,64):
            self.column = 63 - page              
            self.write_cmd(0x00 + (self.column & 0x0f))
            self.write_cmd(0x10 + (self.column >> 4))
            for num in range(0,16):
                self.write_data(self.buffer[page*16+num])


# ---- variable definition ----
global Multiply           # variable for multiply or divide
Multiply = True
global Factor             # factor for multiply/divide
Factor = 5
global TachoInPeriod      # time of the incoming tachosignal period in ms
global TachoOutPeriod     # time of the outgoing tachosignal period in ms
global FirstReading       # true when first reading of the falling edge is needed (timer start)
FirstReading = True
global OutputEnable       # enables the outgoing tachosignal
OutputEnable = False
global start              # start of timer
global RunCore1           # flag if core1 is already running
RunCore1 = False
global DispRow1           # Display Row1
global DispRow2           # Display Row2
global DispRow3           # Display Row3
global DispMul            # Display (Row2) Mul/Div
DispMul = "x"
global DelayTime_ms       # delay time for shut off output (ms)
DelayTime_ms = 3000
global FileName
FileName = "Settings.txt" # define filename for saving the settings into it
global PulsePerRevolution
PulsePerRevolution = 2

# use the inbuilt LED as a file activity light
LED_FileWrite = machine.Pin(25,machine.Pin.OUT)

# writes whatever is passed to it to the file     
def WriteFile(passed):
    LED_FileWrite.value(1)   #indicate writing to file, so don't power off
    log = open(FileName,"a") #open in append - creates if not existing, will append if it exists
    log.write(passed)
    log.close()
    LED_FileWrite.value(0)

# reading the settings from file into variables
def ReadSettings():
    global Multiply
    global Factor
    global DelayTime_ms
    global PulsePerRevolution
    
    try:    
        file = open(FileName, "r")
        Multiply = bool(file.readline().rstrip("\r\n"))
        Factor = int(file.readline().rstrip("\r\n"))
        DelayTime_ms = int(file.readline().rstrip("\r\n"))
        PulsePerRevolution = int(file.readline().rstrip("\r\n"))
        file.close()
    except OSError:
        z=0
    

# ---- Menu for settings ---- 
def Menu():
    global OutputEnable
    global FirstReading
    global DispRow1
    global DispRow2
    global DispRow3
    global DispMul
    global Multiply
    global Factor
    global DelayTime_ms
    global FileName
    global PulsePerRevolution
    
    OutputEnable = False                         # shut off output (core1)
    FirstReading = True                          # allow a new "first reading"

    # menu level0 (multiply/divide)
    MenuLooper = 0
    while MenuLooper == 0:
        DispRow2 = DispMul + str(Factor)
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        time.sleep_ms(500)                       # Blink-Delay 
        DispRow2 = " " + str(Factor)
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        if keyB.value() == 0:
            if Multiply == True:
                Multiply = False                 # set "divide mode"
                DispMul = "/"                   
                DispRow2 = DispMul + str(Factor)
            elif Multiply == False:
                Multiply = True                  # set "multiply mode"
                DispMul = "x"
                DispRow2 = DispMul + str(Factor)
            while keyB.value() == 0:
                z=0 #NOP                         # debounce
        if keyA.value() == 0:
            MenuLooper = MenuLooper + 1          # activate next menu level
            while keyA.value() ==0:
                z=0                              # debounce
    # menu level1 (factor)
    while MenuLooper == 1:
        DispRow2 = DispMul + str(Factor)
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        time.sleep_ms(200)                       # Blink-Delay
        DispRow2 = DispMul + " "
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        if keyB.value() == 0:
            Factor = Factor + 1                  # increment factor
            if Factor > 16:                      # check if factor > 16 to set it back to 1
                Factor = 1  
            DispRow2 = DispMul + str(Factor)
            while keyB.value() == 0:
                z=0 #NOP                         # debounce
        if keyA.value() == 0:
            MenuLooper = MenuLooper + 1          # next menu level
            while keyA.value() ==0:
                z=0                              # debounce
    #menu level2-5 (delay time)
    while 2 <= MenuLooper <= 5:
        ZahlString = str(DelayTime_ms)           # convert DelayTime_ms into string parts
        Tausender = ZahlString[0]                # for possibility to change digit by digit
        try:
            Hunderter = ZahlString[1]
        except IndexError:
            Hunderter = "0"
        try:   
            Zehner = ZahlString[2]
        except IndexError:
            Zehner = "0"
        try:
            Einer = ZahlString[3]
        except IndexError:
            Einer = Tausender                    # set "Einer", which was read as "Tausender" because ZahlString was 0 (too less digits)
            Tausender = "0"                        
       
        DispRow2 = "Delay " + Tausender + Hunderter + Zehner + Einer + "ms"
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        time.sleep_ms(180)                       # Blink-Delay
        if MenuLooper == 2:
            DispRow2 = "Delay " + Tausender + Hunderter + Zehner + " " + "ms"    # blinking "Einer"
        elif MenuLooper == 3:
            DispRow2 = "Delay " + Tausender + Hunderter + " " + Einer + "ms"     # blinking "Zehner"
        elif MenuLooper == 4:
            DispRow2 = "Delay " + Tausender + " " + Zehner + Einer + "ms"        # blinking "Hunderter"
        elif MenuLooper == 5:
            DispRow2 = "Delay " + " " + Hunderter + Zehner + Einer + "ms"        # blinking "Tausender"
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        if keyB.value() == 0:
            if MenuLooper == 2:
                Zahl = int(Einer) + 1             # increment digit
                if Zahl > 9:                      # check if digit > 9 to set it back to 0
                    Zahl = 0
                Einer = str(Zahl)
            elif MenuLooper == 3:
                Zahl = int(Zehner) + 1    
                if Zahl > 9:
                    Zahl = 0
                Zehner = str(Zahl)
            elif MenuLooper == 4:
                Zahl = int(Hunderter) + 1    
                if Zahl > 9:
                    Zahl = 0
                Hunderter = str(Zahl)
            elif MenuLooper == 5:
                Zahl = int(Tausender) + 1    
                if Zahl > 9:
                    Zahl = 0
                Tausender = str(Zahl)
            ZahlString = Tausender + Hunderter + Zehner + Einer
            DelayTime_ms = int(ZahlString)
            DispRow2 = DispMul + str(DelayTime_ms)
            while keyB.value() == 0:
                z=0 #NOP                          # debounce
        if keyA.value() == 0:
            MenuLooper = MenuLooper + 1           # next menu level
            while keyA.value() ==0:
                z=0                               # debounce
    while MenuLooper == 6:
        DispRow2 = "Puls/Rev " + str(PulsePerRevolution)
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        time.sleep_ms(200)                       # Blink-Delay
        DispRow2 = "Puls/Rev " + "  "
        OLED.fill(0x0000)
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.show()
        if keyB.value() == 0:
            PulsePerRevolution = PulsePerRevolution + 1    # increment Puls/Rev
            if PulsePerRevolution > 16:                    # check if Puls/Rev > 16 to set it back to 1
                PulsePerRevolution = 1  
            DispRow2 = "Puls/Rev " + str(PulsePerRevolution)
            while keyB.value() == 0:
                z=0 #NOP                         # debounce
        if keyA.value() == 0:
            MenuLooper = MenuLooper + 1          # next menu level
            while keyA.value() ==0:
                z=0                              # debounce            
                         
# safe settings to file
    while MenuLooper == 7:
        try:
            os.remove(FileName)                   # delete existing file if exists
        except OSError:
            z=0   #NOP                            # do nothing when file doesn't exist (errorhandler)
        StringToWrite = str(Multiply)+ "\r\n"             # add a line feed/carriage return to ensure each number is on its own line in the file
        WriteFile(StringToWrite)                          # safe Multiply
        StringToWrite = str(Factor)+ "\r\n"
        WriteFile(StringToWrite)                          # safe Factor
        StringToWrite = str(DelayTime_ms)+ "\r\n"
        WriteFile(StringToWrite)                          # safe DelayTime        
        StringToWrite = str(PulsePerRevolution)+ "\r\n"
        WriteFile(StringToWrite)                          # safe PulsePerRevolution
        # safescreen
        DispRow1 = "Settings"
        DispRow2 = "SAFED"
        DispRow3 = FileName 
        OLED.fill(0x0000)                         # clear screen
        OLED.text(DispRow1,4,8,OLED.white)
        OLED.text(DispRow2,4,20,OLED.white)
        OLED.text(DispRow3,4,50,OLED.white)
        OLED.show()
        time.sleep_ms(1000)                       # delay for safescreen
        MenuLooper = MenuLooper + 1
#    while MenuLooper > 6:
#        pass    # NOP -> EXIT Menu    

# ----core1 task (calculate and generate outgoing tachosignal when activated trough core0)----
# read OutputEnable, calculate TachoOutPeriod, generate tacho output
def core_task():
    global TachoOutPeriod
    global RunCore1
    global Multiply
    global TachoInPeriod
    global Factor

    while OutputEnable == True:                             # output on?
        if Multiply == True:                                # multiply (faster output signal)
            TachoOutPeriod = int(TachoInPeriod / Factor)
        else:                                               # divide (slower output signal)
            TachoOutPeriod = int(TachoInPeriod * Factor)
        OutTachoDuet.toggle()                               # toggle tacho output
        time.sleep_ms(int(TachoOutPeriod / 2))              # wait a "half period" to toggle output again
    RunCore1 = False
    _thread.exit()                                          # shut off core1

OutTachoDuet=Pin(21, Pin.OPEN_DRAIN, value=1)               # tachosignal output (open drain) for DUET

# ----core0 task (input reading, measuring incoming tachosignal)----
# interrupt handler at falling edge of tacho input, measures tacho input, sets OutputEnable, starts/stops core1
def InTacho_readTask(pin):         # interrupt from tacho input
    global start
    global FirstReading
    global OutputEnable
    global RunCore1
    global TachoInPeriod

    if FirstReading == True:       # first falling edge of incoming signal?
        start = time.ticks_ms()    # get millisecond counter
        FirstReading = False       # first reading (timerstart) done.
    else:
        TachoInPeriod = time.ticks_diff(time.ticks_ms(), start)  # compute time difference between falling edges
        if TachoInPeriod == 0:
            TachoInPeriod = 0
        start = time.ticks_ms()                                  # start a new timer
        OutputEnable = True                                      # activate tacho output
        if RunCore1 == False:                                    # core1 already running?
            RunCore1 = True
            _thread.start_new_thread((core_task),())             # start core1

InTachoFlowmeter=Pin(20, Pin.IN, Pin.PULL_UP)                              # flowmeter tachosignal, with pullup(switch to GND)
InTachoFlowmeter.irq(trigger=Pin.IRQ_FALLING, handler=InTacho_readTask)    # interrupt on falling edge


#---- main task ----  
# startscreen
DispRow1 = "TachoMod V4.0"
DispRow2 = " "
DispRow3 = "by cosmowave"

OLED = OLED_1inch3()
OLED.fill(0x0000)                     # clear screen
OLED.text(DispRow1,10,8,OLED.white)
OLED.text(DispRow2,22,20,OLED.white)
OLED.text(DispRow3,30,50,OLED.white)
OLED.show()

time.sleep_ms(2400)                   # delay for startscreen

# read settings 
ReadSettings()
print(Multiply)
print(Factor)
print(DelayTime_ms)
print(PulsePerRevolution)

# ready screen
OLED.fill(0x0000)
OLED.text(DispRow1,10,8,OLED.white)
DispRow2 = "READY !"
OLED.text(DispRow1,10,8,OLED.white)
OLED.text(DispRow2,22,20,OLED.white)
OLED.show()

#---- Main loop ----
# check if no tacho input since 3seconds to switch off output, Menu entry only possible while no input signal detected

keyA = Pin(15,Pin.IN,Pin.PULL_UP)    # define keyA (Menu entry)
keyB = Pin(17,Pin.IN,Pin.PULL_UP)    # define keyB (change value)
while True:
    global start
    global OutputEnable
    global FirstReading
    global TachoInPeriod
    global TachoOutPeriod
    global DispRow1
    global DispRow2
    global DispRow3
    global DispMul
    global PulsePerRevolution

    while OutputEnable == True:
        DispRow1 = "IN  " + str(TachoInPeriod) + " / " + str(int(1/TachoInPeriod*1000*60/PulsePerRevolution))
        DispRow2 = DispMul + str(Factor) + "   ms    rpm"
        if int(1/TachoOutPeriod*1000*60) == 0:
            TachoOutPeriod = 1
        DispRow3 = "OUT "+ str(TachoOutPeriod)+" / "+ str(int(1/TachoOutPeriod*1000*60/PulsePerRevolution))

        OLED.fill(0x0000)                     # clear oled screen
        OLED.rect(0,0,128,22,OLED.white)
        OLED.rect(0,42,128,22,OLED.white)
        OLED.text(DispRow1,4,8,OLED.white)
        OLED.text(DispRow2,4,28,OLED.white)
        OLED.text(DispRow3,4,50,OLED.white)
        OLED.show()

    # timeout check
        if time.ticks_diff(time.ticks_ms(), start) > DelayTime_ms:   # more than DelayTime without tacho input
            OutputEnable = False                                     # shut off output (core1)
            FirstReading = True                                      # allow a new "first reading"

            # "no input" display
            DispRow1 = "No IN signal"
            DispRow2 = DispMul + str(Factor)+"   " + str(DelayTime_ms) + "ms"
            DispRow3 = "Output off"
            OLED.fill(0x0000)                     # clear oled screen
            OLED.rect(0,0,128,22,OLED.white)
            OLED.rect(0,42,128,22,OLED.white)
            OLED.text(DispRow1,4,8,OLED.white)
            OLED.text(DispRow2,4,28,OLED.white)
            OLED.text(DispRow3,4,50,OLED.white)
            OLED.show()
    # menu
    if keyA.value() == 0:
            Menu()
            # screen after leaving menu
            DispRow1 = "No IN signal"
            DispRow2 = DispMul + str(Factor) +"   " + str(DelayTime_ms) + "ms"
            DispRow3 = "Output off"
            OLED.fill(0x0000)                     # clear oled screen
            OLED.rect(0,0,128,22,OLED.white)
            OLED.rect(0,42,128,22,OLED.white)
            OLED.text(DispRow1,4,8,OLED.white)
            OLED.text(DispRow2,4,28,OLED.white)
            OLED.text(DispRow3,4,50,OLED.white)
            OLED.show()